@@ -1,17 +1,17 @@
 import abc
 import asyncio
+import cProfile
 import functools
 import ipaddress
 import json
 import logging
+import pstats
 import re
+import string
 import types
 import typing
-import string
 import uuid
 from email.utils import parseaddr
-import cProfile
-import pstats
 
 import acme.jws
 import acme.messages
@@ -25,23 +25,17 @@ from cryptography import x509
 from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.asymmetric import rsa, ec
 
+import acmetk.util
 from acmetk import models
 from acmetk.client import CouldNotCompleteChallenge, AcmeClientException, AcmeClient
 from acmetk.database import Database
 from acmetk.models import messages
 from acmetk.server import ChallengeValidator
-from acmetk.server.external_account_binding import AcmeEAB
-from acmetk.server.management import AcmeManagement
+from acmetk.server.external_account_binding import AcmeEABMixin
+from acmetk.server.management import AcmeManagementMixin
 from acmetk.server.routes import routes
-from acmetk.util import (
-    url_for,
-    generate_cert_from_csr,
-    names_of,
-    forwarded_url,
-    pem_split,
-    ConfigurableMixin,
-)
 from acmetk.version import __version__
+from acmetk.plugin_base import PluginRegistry
 
 logger = logging.getLogger(__name__)
 
@@ -68,15 +62,16 @@ class AcmeResponse(web.Response):
         )
 
 
-class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
+class AcmeServerBase(AcmeEABMixin, AcmeManagementMixin, abc.ABC):
     """Base class for an ACME compliant server.
 
-    Implementations must set the :attr:`config_name` attribute, so that the CLI script knows which
-    configuration option corresponds to which server class.
+    Implementations must also be registered with the plugin registry via
+    :meth:`~acmetk.plugin_base.PluginRegistry.register_plugin`, so that the CLI script knows which configuration
+    option corresponds to which server class.
     """
 
-    config_name: str
-    """The string that maps to the server implementation inside configuration files."""
+    ORDERS_LIST_CHUNK_LEN = 10
+    """Number of order links to include per request."""
 
     SUPPORTED_JWS_ALGORITHMS = (
         josepy.jwa.RS256,
@@ -115,8 +110,6 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
     better than nothing, but  accepts names ending with -
     """
 
-    subclasses = []
-
     def __init__(
         self,
         *,
@@ -302,7 +295,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
 
         return AcmeResponse(
             self._issue_nonce(),
-            url_for(request, "directory"),
+            acmetk.util.url_for(request, "directory"),
             *args,
             **kwargs,
             text=text,
@@ -376,7 +369,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
 
         sig = jws.signature.combined
 
-        if sig.url != str(forwarded_url(request)):
+        if sig.url != str(acmetk.util.forwarded_url(request)):
             raise acme.messages.Error.with_code("unauthorized")
 
         if sig.alg not in self.SUPPORTED_JWS_ALGORITHMS:
@@ -393,6 +386,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
             raise acme.messages.Error.with_code("malformed")
 
         if key_auth:
+            # check whether key was supplied - josepy.errors.Error: No key found - malformed
             if not jws.verify(sig.jwk):
                 raise acme.messages.Error.with_code("unauthorized")
             else:
@@ -400,11 +394,16 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
         elif sig.kid:
             account_id = yarl.URL(sig.kid).name
 
-            if url_for(request, "accounts", account_id=account_id) != sig.kid:
+            if (
+                acmetk.util.url_for(request, "accounts", account_id=account_id)
+                != sig.kid
+            ):
                 """Bug in the dehydrated client, accepted by boulder, so we accept it too.
                 Dehydrated puts .../new-account/{kid} into the request signature, instead of
                 .../accounts/{kid}."""
-                kid_new_account_route = yarl.URL(url_for(request, "new-account"))
+                kid_new_account_route = yarl.URL(
+                    acmetk.util.url_for(request, "new-account")
+                )
                 kid_new_account_route = kid_new_account_route.with_path(
                     kid_new_account_route.path + "/" + account_id
                 )
@@ -447,9 +446,13 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
             jws, account = await self._verify_request(request, session)
         except acme.messages.Error:
             data = await request.text()
-            jws = acme.jws.JWS.json_loads(
-                data
-            )  # TODO: raise acme error on deserialization error
+
+            try:
+                jws = acme.jws.JWS.json_loads(data)
+            except josepy.errors.DeserializationError:
+                raise acme.messages.Error.with_code(
+                    "malformed", detail="The request does not contain a valid JWS."
+                )
             account = None
 
         try:
@@ -578,11 +581,11 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
         :return: The directory object.
         """
         directory = {
-            "newAccount": url_for(request, "new-account"),
-            "newNonce": url_for(request, "new-nonce"),
-            "newOrder": url_for(request, "new-order"),
-            "revokeCert": url_for(request, "revoke-cert"),
-            "keyChange": url_for(request, "key-change"),
+            "newAccount": acmetk.util.url_for(request, "new-account"),
+            "newNonce": acmetk.util.url_for(request, "new-nonce"),
+            "newOrder": acmetk.util.url_for(request, "new-order"),
+            "revokeCert": acmetk.util.url_for(request, "revoke-cert"),
+            "keyChange": acmetk.util.url_for(request, "key-change"),
             "meta": {},
         }
 
@@ -643,7 +646,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
                         request,
                         account.serialize(request),
                         headers={
-                            "Location": url_for(
+                            "Location": acmetk.util.url_for(
                                 request, "accounts", account_id=str(account.account_id)
                             )
                         },
@@ -675,7 +678,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
                         serialized,
                         status=201,
                         headers={
-                            "Location": url_for(
+                            "Location": acmetk.util.url_for(
                                 request, "accounts", account_id=str(account_id)
                             )
                         },
@@ -762,7 +765,9 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
             request,
             serialized,
             status=201,
-            headers={"Location": url_for(request, "order", id=str(order_id))},
+            headers={
+                "Location": acmetk.util.url_for(request, "order", id=str(order_id))
+            },
         )
 
     @routes.post("/authz/{id}", name="authz")
@@ -891,18 +896,49 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
 
     @routes.post("/orders/{id}", name="orders")
     async def orders(self, request):
-        """Handler that retrieves the account's orders list.
+        """Handler that retrieves the account's chunked orders list.
 
         `7.1.2.1.  Orders List <https://tools.ietf.org/html/rfc8555#section-7.1.2.1>`_
 
-        :return: An object with key *orders* that holds the account's orders list.
+        :return: An object with key *orders* that holds a chunk of the account's orders list.
         """
         async with self._session(request) as session:
             jws, account = await self._verify_request(
                 request, session, post_as_get=True
             )
+            try:
+                cursor = int(request.query.get("cursor", 0))
+                orders = await self._db.get_orders_list(
+                    session, account.account_id, self.ORDERS_LIST_CHUNK_LEN, cursor
+                )
+            except ValueError:
+                raise web.HTTPBadRequest(text="Cursor must be an integer >= 0.")
 
-            return self._response(request, {"orders": account.orders_list(request)})
+            if len(orders) == 0:
+                raise web.HTTPNotFound(
+                    text="No orders found. Try a lower cursor value or create some orders first."
+                )
+
+            """The next two lines ensure that the extra order we query to see if there more orders is not returned
+            to the client. If there are no more orders after the current cursor, then return all of them."""
+            more_orders = len(orders) == (self.ORDERS_LIST_CHUNK_LEN + 1)
+            orders = orders[:-1] if more_orders else orders
+
+            return self._response(
+                request,
+                {
+                    "orders": [
+                        order.url(request)
+                        for order in orders
+                        if order.status == models.OrderStatus.PENDING
+                    ]
+                },
+                links=[
+                    f'<{acmetk.util.next_url(account.orders_url(request), cursor)}>; rel="next"'
+                ]
+                if more_orders
+                else [],
+            )
 
     async def _validate_order(
         self, request, session
@@ -995,7 +1031,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
         return self._response(
             request,
             serialized,
-            headers={"Location": url_for(request, "order", id=order_id)},
+            headers={"Location": acmetk.util.url_for(request, "order", id=order_id)},
         )
 
     @routes.post("/certificate/{id}", name="certificate")
@@ -1074,7 +1110,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
 
             key_change = messages.KeyChange.json_loads(inner_jws.payload)
 
-            if key_change.account != url_for(
+            if key_change.account != acmetk.util.url_for(
                 request, "accounts", account_id=str(account.account_id)
             ):
                 """7.  Check that the "account" field of the keyChange object contains
@@ -1115,7 +1151,7 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
                 request,
                 serialized,
                 headers={
-                    "Location": url_for(
+                    "Location": acmetk.util.url_for(
                         request, "accounts", account_id=str(account.account_id)
                     )
                 },
@@ -1228,11 +1264,10 @@ class AcmeServerBase(AcmeEAB, AcmeManagement, ConfigurableMixin, abc.ABC):
             return response
 
 
+@PluginRegistry.register_plugin("ca")
 class AcmeCA(AcmeServerBase):
     """ACME compliant Certificate Authority."""
 
-    config_name = "ca"
-
     def __init__(self, *, cert, private_key, **kwargs):
         super().__init__(**kwargs)
 
@@ -1274,7 +1309,9 @@ class AcmeCA(AcmeServerBase):
         async with self._session(request) as session:
             order = await self._db.get_order(session, account_id, order_id)
 
-            cert = generate_cert_from_csr(order.csr, self._cert, self._private_key)
+            cert = acmetk.util.generate_cert_from_csr(
+                order.csr, self._cert, self._private_key
+            )
             order.certificate = models.Certificate(
                 status=models.CertificateStatus.VALID, cert=cert
             )
@@ -1369,7 +1406,9 @@ class AcmeRelayBase(AcmeServerBase):
 
             return self._response(
                 request,
-                text=certificate.full_chain,
+                body=certificate.full_chain.encode(),
+                links=None,
+                content_type="application/pem-certificate-chain",
             )
 
     # @routes.post("/revoke-cert", name="revoke-cert")
@@ -1422,7 +1461,7 @@ class AcmeRelayBase(AcmeServerBase):
         :param order_ca: The remote CA's order object
         """
         full_chain = await self._client.certificate_get(order_ca)
-        certs = pem_split(full_chain)
+        certs = acmetk.util.pem_split(full_chain)
 
         if len(certs) < 2:
             logger.info(
@@ -1440,6 +1479,7 @@ class AcmeRelayBase(AcmeServerBase):
             order.status = models.OrderStatus.VALID
 
 
+@PluginRegistry.register_plugin("broker")
 class AcmeBroker(AcmeRelayBase):
     """Server that relays requests to a remote CA employing a "broker" model.
 
@@ -1449,8 +1489,6 @@ class AcmeBroker(AcmeRelayBase):
     the :class:`AcmeProxy` class should be used instead.
     """
 
-    config_name = "broker"
-
     async def handle_order_finalize(self, request, account_id: str, order_id: str):
         """Method that handles the actual finalization of an order.
 
@@ -1474,7 +1512,9 @@ class AcmeBroker(AcmeRelayBase):
             order = await self._db.get_order(session, account_id, order_id)
 
             try:
-                order_ca = await self._client.order_create(list(names_of(order.csr)))
+                order_ca = await self._client.order_create(
+                    list(acmetk.util.names_of(order.csr))
+                )
                 await self._client.authorizations_complete(order_ca)
                 finalized = await self._client.order_finalize(order_ca, order.csr)
                 await self.obtain_and_store_cert(order, finalized)
@@ -1502,6 +1542,7 @@ class AcmeBroker(AcmeRelayBase):
             await session.commit()
 
 
+@PluginRegistry.register_plugin("proxy")
 class AcmeProxy(AcmeRelayBase):
     """Server that relays requests to a remote CA employing a "proxy" model.
 
@@ -1509,8 +1550,6 @@ class AcmeProxy(AcmeRelayBase):
     the possibility to show errors to the end user as they occur at the remote CA.
     """
 
-    config_name = "proxy"
-
     # @routes.post("/new-order", name="new-order")
     async def new_order(self, request):
         """Handler that creates a new order.
@@ -1549,7 +1588,9 @@ class AcmeProxy(AcmeRelayBase):
             request,
             serialized,
             status=201,
-            headers={"Location": url_for(request, "order", id=str(order_id))},
+            headers={
+                "Location": acmetk.util.url_for(request, "order", id=str(order_id))
+            },
         )
 
     async def _complete_challenges(self, request, account_id, order_id):
@@ -1642,7 +1683,7 @@ class AcmeProxy(AcmeRelayBase):
         return self._response(
             request,
             serialized,
-            headers={"Location": url_for(request, "order", id=order_id)},
+            headers={"Location": acmetk.util.url_for(request, "order", id=order_id)},
         )
 
     async def handle_order_finalize(self, request, account_id: str, order_id: str):
